/*
**
**  WAY (Where Are You ?)
**
**  dbase.c		database module.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   890728  Roberto Bagnara  First working version
**
**
**  Description:
**
**  this module contains the database procedures used by the name server.
**  It provides functions to establish associations between items, each
**  item being identified by its name. In other words it allows to store
**  and retrieve ordered triples (S, A, I) where S and A are items, while
**  I characterizes the association between S and A.
**
*/

/*
**      Include files.
**      --------------
*/

#include "way.h"
#include <strings.h>
#include "info.h"
#include "dbase.h"

/*
**      External functions.
**      -------------------
*/

#ifdef __STDC__
extern char *xmalloc(int size);
extern void xfree();
#else
extern char *xmalloc();
extern void xfree();
#endif

/*
**      These are the roots of our trees.
**      ---------------------------------
*/

static TREE_PTR s_root, a_root;


/*
**
**	The tree search function.
**	-------------------------
**
**	
**	TREE_PTR tree_search(const TREE_PTR root, const char *name)
**
**  Parameters:	root		the tree root;
**		name		the name to look for in the tree.
**
**  Return:			pointer to a tree node or NULL_TREE_PTR.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Search the tree whose root is "root" for a node with name field set
**	to "name". If the search succeeds a pointer to that element is returned,
**	otherwise NULL_TREE_PTR is returned.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
static TREE_PTR tree_search(const TREE_PTR root, const char *name)
#else
static TREE_PTR tree_search(root, name)
TREE_PTR root;
char *name;
#endif
{
    if (root == NULL_TREE_PTR)
	return NULL_TREE_PTR;

    else {

	register int comp;

	comp = strcmp(name, root->name);

	if (comp == 0)
	    return root;
	else if (comp < 0)
	    return tree_search(root->left, name);
	else
	    return tree_search(root->right, name);
    }
}


/*
**
**	The tree insert function.
**	-------------------------
**
**	
**	TREE_PTR tree_insert(TREE_PTR *root, const char *name)
**
**  Parameters:	root		the tree root;
**		name		the name to insert in the tree.
**
**  Return:			pointer to a tree node.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**  Inserts a node with name field set to "name" in the tree whose root
**  is "root", if it is not already present. This function returns a pointer
**  to the (new or not) tree node concerned.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
static TREE_PTR tree_insert(TREE_PTR *root, const char *name)
#else
static TREE_PTR tree_insert(root, name)
TREE_PTR *root;
char *name;
#endif
{
    if (*root == NULL_TREE_PTR) {

	/* A node identified by name is not present, insert it. */

	*root = (TREE_PTR) xmalloc(sizeof(TREE_NODE));

	/* Initialize pointers to childs and to the association list. */

	(*root)->left = (*root)->right = NULL_TREE_PTR;
	(*root)->assoc = NULL_LIST_PTR;

	/* Setup name and pointer to it. */

	(*root)->name = strcpy(xmalloc(strlen(name) + 1), name);

	/* Return a pointer to the new node. */

	return *root;

    } else {

	register int comp;

        comp = strcmp(name, (*root)->name);

        if (comp == 0)
	    return *root;
	else if (comp < 0)
	    return tree_insert(&((*root)->left), name);
	else
	    return tree_insert(&((*root)->right), name);
    }
}


/*
**
**	The tree remove function.
**	-------------------------
**
**	
**	void tree_remove(TREE_PTR *root, const char *name)
**
**  Parameters:	root		the tree root;
**		name		the name to remove from the tree.
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**  Remove node with name field set to "name" from the tree whose root
**  is "root", if it is present.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

/*
**      This is to handle the case where the node to remove has both childs.
**      --------------------------------------------------------------------
*/

#ifdef __STDC__
static void tree_rem(TREE_PTR *node, TREE_PTR *substitute)
#else
static void tree_rem(node, substitute)
TREE_PTR *node, *substitute;
#endif
{
    if ((*substitute)->right == NULL_TREE_PTR) {

	/* The rightmost child has been found, copy the relevant stuff. */

	(void) strcpy((*node)->name, (*substitute)->name);
	(*node)->assoc = (*substitute)->assoc;

	/* Return back a pointer to node to be discarded. */

	*node = *substitute;

	/* The rightmost child is substituted with its left child. */

	*substitute = (*substitute)->left;
    } else

	/* Continue descending on the right side. */

	tree_rem(node, &((*substitute)->right));
}

#ifdef __STDC__
static void tree_remove(TREE_PTR *root, const char *name)
#else
static void tree_remove(root, name)
TREE_PTR *root;
char *name;
#endif
{
    /* The node doesn't exist, return silently. */

    if (*root == NULL_TREE_PTR)
	return;

    else {

	register int comp;

	comp = strcmp(name, (*root)->name);

	if (comp == 0) {

	    TREE_PTR tmp;

	    /* The node to be deleted has been found. */

	    tmp = *root;

	    /* If it hasn't a right child substitute it with its left one */

	    if ((*root)->right == NULL_TREE_PTR)
		*root = (*root)->left;

	    /* If it hasn't a left child substitute it with its right one */

	    else if (tmp->left == NULL_TREE_PTR)
		*root = (*root)->right;

	    /* If it has both childs substitute it with
	       the rightmost child of its left child. */

	    else
		tree_rem(&tmp, &(tmp->left));

	    /* Release memory */

	    xfree(tmp->name);
	    xfree(tmp);

	} else if (comp < 0)
	    tree_remove(&((*root)->left), name);
	else
	    tree_remove(&((*root)->right), name);
    }
}


/*
**
**	The list insert function.
**	-------------------------
**
**	
**	LIST_PTR list_insert(const TREE_PTR s_node, const TREE_PTR a_node)
**
**  Parameters:	s_node		one end of the link;
**		a_node		the other end of the link.
**
**  Return:			pointer to an association list item.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**  Inserts an association item linking "s_node" to "a_node" into the
**  association lists of "s_node" and "a_node", if it is not already present.
**  This function returns a pointer to the (new or not) association item.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
static LIST_PTR list_insert(const TREE_PTR s_node, const TREE_PTR a_node)
#else
static LIST_PTR list_insert(s_node, a_node)
TREE_PTR s_node, a_node;
#endif
{
    register LIST_PTR tmp;

    /* This is the head of the associoation list of s_node. */

    tmp = s_node->assoc;

    /* See if an association linking s_node to a_node already exists. */

    while ((tmp != NULL_LIST_PTR) && ((tmp->s_node != s_node) ||
				      (tmp->a_node != a_node)))
	tmp = tmp->s_next;

    if (tmp == NULL_LIST_PTR) {

	/* A list item associating s_node to a_node
	   is not present, insert it.
	*/

	tmp = (LIST_PTR) xmalloc(sizeof(LIST_NODE));

	/* Initialize pointers to associated nodes. */

	tmp->s_node = s_node;
	tmp->a_node = a_node;

	/* Insert the new association item on the
	   s_node and a_node association lists.
	*/

	tmp->s_next = s_node->assoc;
	tmp->a_next = a_node->assoc;
	s_node->assoc = a_node->assoc = tmp;
    }

    /* Return a pointer to the new node */

    return tmp;
}


/*
**
**	The list remove function.
**	-------------------------
**
**	
**	void list_remove(const TREE_PTR s_node, const TREE_PTR a_node)
**
**  Parameters:	s_node		one end of the link;
**		a_node		the other end of the link.
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**  Removes the association item linking "s_node" to "a_node" from the
**  association lists of "s_node" and "a_node", if it exists.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
static void list_remove(const TREE_PTR s_node, const TREE_PTR a_node)
#else
static void list_remove(s_node, a_node)
TREE_PTR s_node, a_node;
#endif
{
    register LIST_PTR scan1, scan2, prev;

    /* See if an association linking s_node to a_node exists
       on the s_node association list, if it does remove it.
    */

    prev = NULL_LIST_PTR;
    scan1 = s_node->assoc;
    while ((scan1 != NULL_LIST_PTR) && ((scan1->s_node != s_node) ||
					(scan1->a_node != a_node))) {
	prev = scan1;
	scan1 = scan1->s_next;
    }

    if (scan1 != NULL_LIST_PTR) {

	/* Found. */

	if (prev == NULL_LIST_PTR)

	    /* It is the first list item. */

	    s_node->assoc = scan1->s_next;

	else

	    /* It isn't the first list item. */

	    prev->s_next = scan1->s_next;

	/* Remove it also from the a_node association list. */

	prev = NULL_LIST_PTR;
	scan2 = a_node->assoc;
	while (scan2 != scan1) {
	    prev = scan2;
	    scan2 = scan2->a_next;
	}

	if (prev == NULL_LIST_PTR)
	    a_node->assoc = scan2->a_next;
	else
	    prev->a_next = scan2->a_next;

	/* Release memory. */

	xfree(scan2);
    }
}


/*
**
**	The database initialization procedure.
**	--------------------------------------
**
**	
**	void db_init(void)
**
**  Parameters:	none
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	s_root		set to NULL_TREE_PTR;
**		a_root		set to NULL_TREE_PTR.
**
**  Description:
**
**	Initializes the data structures used in this module. The trees roots
**	are just set to NULL_TREE_PTR. This means that db_init() is intended
**	only as an INIT procedure, NOT a RESET one, as any space previously
**	allocated is not released. There is no protection on that, user
**	modules are simply expected to call db_init ONCE.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
void db_init(void)
#else
void db_init()
#endif
{
    s_root = a_root = NULL_TREE_PTR;
}


/*
**
**	The database S search procedure.
**	--------------------------------
**
**	
**	LIST_PTR db_search_s(const char *name)
**
**  Parameters:	name		name of the S item to look for.
**
**  Return:			pointer to a list node or NULL_LIST_PTR.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Search the S tree for a node with name field set to "name".
**	If the search succeeds a pointer to the node association list head
**	is returned, otherwise NULL_LIST_PTR is returned.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
LIST_PTR db_search_s(const char *name)
#else
LIST_PTR db_search_s(name)
char *name;
#endif
{
    TREE_PTR node;

    if ((node = tree_search(s_root, name)) == NULL_TREE_PTR)
	return NULL_LIST_PTR;
    else
	return node->assoc;
}


/*
**
**	The database A search procedure.
**	--------------------------------
**
**	
**	LIST_PTR db_search_a(const char *name)
**
**  Parameters:	name		name of the A item to look for.
**
**  Return:			pointer to a list node or NULL_LIST_PTR.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Search the A tree for a node with name field set to "name".
**	If the search succeeds a pointer to the node association list
**	head is returned, otherwise NULL_LIST_PTR is returned.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
LIST_PTR db_search_a(const char *name)
#else
LIST_PTR db_search_a(name)
char *name;
#endif
{
    TREE_PTR node;

    if ((node = tree_search(a_root, name)) == NULL_TREE_PTR)
	return NULL_LIST_PTR;
    else
	return node->assoc;
}


/*
**
**	The database association procedure.
**	-----------------------------------
**
**	
**	void db_associate(const char *s_name, const char *a_name, const INFO info)
**
**  Parameters:	s_name		name of the S item to associate;
**		a_name		name of the A item to associate;
**		info		association additional informations.
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Establish an association between "s_name" and "a_name" characterized
**	by the informations passed by "info", if it does not exist already.
**	Even if it does already exist the new additional informations
**	passed by "info" override the old ones.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
void db_associate(const char *s_name, const char *a_name, const INFO info)
#else
void db_associate(s_name, a_name, info)
char *s_name, *a_name;
INFO info;
#endif
{
    TREE_PTR s_node, a_node;
    LIST_PTR link;

    /* Insert s_name in the s_root tree. */

    s_node = tree_insert(&s_root, s_name);

    /* Insert a_name in the a_root tree. */

    a_node = tree_insert(&a_root, a_name);

    /* Associate them. */

    link = list_insert(s_node, a_node);

    /* Copy the additional informations. */

    link->info = info;
}


/*
**
**	The database dissociation procedure.
**	------------------------------------
**
**	
**	void db_dissociate(const char *s_name, const char *a_name)
**
**  Parameters:	s_name		name of the S item to dissociate;
**		a_name		name of the A item to dissociate;
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Disestablish the association between "s_name" and "a_name" , if it
**	exists.
**
**  History:
**
**	28 Jul 89	Written. (RB)
**
*/

#ifdef __STDC__
void db_dissociate(const char *s_name, const char *a_name)
#else
void db_dissociate(s_name, a_name)
char *s_name, *a_name;
#endif
{
    TREE_PTR s_node, a_node;

    /* Look for s_name in the s_root tree */

    if ((s_node = tree_search(s_root, s_name)) == NULL_TREE_PTR)
	return;

    /* Look for a_name in the a_root tree */

    if ((a_node = tree_search(a_root, a_name)) == NULL_TREE_PTR)
	return;

    /* Dissociate them. */

    list_remove(s_node, a_node);

    /* If the association list of s_node is empty remove s_node from s_root tree. */

    if (s_node->assoc == NULL_LIST_PTR)
	tree_remove(&s_root, s_name);

    /* If the association list of a_node is empty remove a_node from a_root tree. */

    if (a_node->assoc == NULL_LIST_PTR)
	tree_remove(&a_root, a_name);
}

/*
**	The following code is absolutely provisional !!!
**	------------------------------------------------
*/

static int stop_visit;

#ifdef __STDC__
static void db_visit(const TREE_PTR root, int (*action)(TREE_PTR node))
#else
static void db_visit(root, action)
TREE_PTR root;
int (*action)();
#endif
{
    if (stop_visit || root == NULL_TREE_PTR)
	return;

    stop_visit = (*action)(root);
    db_visit(root->left, action);
    db_visit(root->right, action);
}

#ifdef __STDC__
void db_visit_s(int (*action)(TREE_PTR node))
#else
void db_visit_s(action)
int (*action)();
#endif
{
    stop_visit = 0;
    db_visit(s_root, action);
}

#ifdef __STDC__
void db_visit_a(int (*action)(TREE_PTR node))
#else
void db_visit_a(action)
int (*action)();
#endif
{
    stop_visit = 0;
    db_visit(a_root, action);
}
