/*
**
**  WAY (Where Are You ?)
**
**  info.h		definitions for registration additional informations.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   890901  Roberto Bagnara  First working version
**  1.0.01   891107  Roberto Bagnara  Two fields added to the additional
**				      information structure: version and
**				      compatibility. They are needed for
**				      version checking. Keyword table
**				      initializer modified accordingly.
**
**
**  Description:
**
**	some definitions concerning the additional informations related to
**	any particular registration are collected here.  At the time of
**	writing (September 1st, 1989), it is not very clear which kind of
**	informations have to be collected and how they should be used.
**
**	What I can foresee is that there will be a certain number of flags
**	to control various things, some numeric informations and possibly
**	some string values. Some of these have already been defined, but
**	nobody can tell what will prove to be necessary in the future.
**
**	In order to deal with this the structures and procedures for info
**	handling have been designed to be as flexible as possible, as you
**	can see below.  This allows for easy addition of new information
**	items, while providing an elegant method to handle them.
**
**				WARNING
**				-------
**
**			IF YOU CHANGE THE DEFINITIONS BELOW,
**		PLEASE MAKE SURE TO PRESERVE THEIR CONSISTENCY !!!
**
*/

/*
**	This is the info structure.
**	---------------------------
*/

typedef struct info {
    long flags;			/* Flags controlling various things. */
    union {
	struct {
	    long version;	/* Current version number. */
	    long compatibility;	/* Compatibility version number. */
	    long package_no;	/* Package number retrieved from server. */
	    long timeout;	/* Preferred timeout retrieved from server. */
	} longs;
	long array[4];		/* To view the longs as an array. */
    } long_info;
} INFO, *INFO_PTR;

/*
**	Define the default info flags.
**	------------------------------
**
**	Means "nosecure", "nostatic", "noname_server", "poll_allowed".
*/

#define DEFAULT_FLAGS 8

/*
**	By default we don't know anything.
**	----------------------------------
*/

#define DEFAULT_VERSION		     65535
#define DEFAULT_COMPATIBILITY		 0
#define DEFAULT_PACKAGE_NO		-1
#define DEFAULT_TIMEOUT		 	 0

/*
**	This macro sets up the info defaults.
**	-------------------------------------
*/

#define STORE_DEFAULT_INFO(info) \
{									\
	(info)->flags = DEFAULT_FLAGS;					\
	(info)->long_info.longs.version = DEFAULT_VERSION;		\
	(info)->long_info.longs.compatibility = DEFAULT_COMPATIBILITY;	\
	(info)->long_info.longs.package_no = DEFAULT_PACKAGE_NO;	\
	(info)->long_info.longs.timeout = DEFAULT_TIMEOUT;		\
}

/*
**	For the keyword type we use the definitions generated
**	by bison for the correspondent tokens.
*/

#include "parsetab.h"

/*
**	Just for clarity (see below).
**	-----------------------------
*/

#define DONT_CARE 0

/*
**	This is the keyword table initializer.
**	--------------------------------------
**
**	This table contains each info keyword with its type and an additional
**	"magic" field. When the keyword type is FLAG "magic" provides the
**	indication of which bits should be set or cleared. If "magic" is
**	positive, its binary representation corresponds to the bits to set.
**	If "magic" is negative, the binary representation of its 2's
**	complement corresponds to the bits to clear. If the keyword type
**	is NUMERIC "magic" is the index to access the correspondent long
**	variable in the info steructure. The STRING type is not implemented
**	yet.
*/

#define INFO_KEYWORDS		\
  { { FLAG,    "secure",          1			}, \
    { FLAG,    "nosecure",       -1			}, \
    { FLAG,    "static",          2			}, \
    { FLAG,    "nostatic",       -2			}, \
    { FLAG,    "name_server",     4			}, \
    { FLAG,    "noname_server",  -4			}, \
    { FLAG,    "poll_allowed",    8			}, \
    { FLAG,    "nopoll_allowed", -8			}, \
    { NUMERIC, "version",         0			}, \
    { NUMERIC, "compatibility",   1			}, \
    { NUMERIC, "package_no",      2			}, \
    { NUMERIC, "timeout",         3			} }

/*
**	Macros to make life easier when handling the info flags.
**	--------------------------------------------------------
*/

#define IS_SECURE(x)       (x & 1)
#define IS_STATIC(x)       (x & 2)
#define IS_NAME_SERVER(x)  (x & 4)
#define IS_POLL_ALLOWED(x) (x & 8)

